@@ -7,6 +7,11 @@ from sklearn.utils.validation import check_is_fitted, check_array, _check_y
 from sklearn.preprocessing import LabelEncoder
 from sklearn.calibration import CalibratedClassifierCV
 
+from assembled.utils.logger import get_logger
+from pandas.util import hash_array
+
+logger = get_logger(__file__)
+
 
 class FakedClassifier(BaseEstimator, ClassifierMixin):
     """A fake classifier that simulates a real classifier from the prediction data of the real classifier.
@@ -14,7 +19,28 @@ class FakedClassifier(BaseEstimator, ClassifierMixin):
         We assume the input passed to init is the same format as training X (will be validated in predict).
         We store the prediction data with an index whereby the index is the hash of an instance.
             Some Assumptions of this:
-                - Simulated Models return the same results for the same input instance
+                TODO: add checks for this
+                    - Simulated Models return the same results for the same input instance
+                    - Input data is only numeric (as the default preprocessor makes sure)
+
+
+        !Warnings!:
+            - If the simulated model returns different results for the same input instance (e.g., as a result of
+            using cross-validation to produce the validation data), we set the prediction values for all duplicates
+            to the first value seen for the duplicates.
+
+        A Remark on Hashing if Duplicates are Present:
+            The hash we are using is consistent between runs of the same INTERPRETER, i.e., Python Version
+            (see https://stackoverflow.com/a/64356731). It is only consistent because we are hashing a tuple of
+            numeric values. This would not work for hashes of strings without changing the code
+            (see https://stackoverflow.com/a/2511075).
+
+            Consistency is required if duplicates are present, because otherwise, in the current implementation,
+            the prediction value selected to represent all duplicates might change and thus the data that is passed
+            to an ensemble method would change.
+
+            TODO: re-implement this, change hash method, or think of different approach to
+                non-restrictive index management
 
     Parameters
     ----------
@@ -46,6 +72,8 @@ class FakedClassifier(BaseEstimator, ClassifierMixin):
         Time the real model took to evaluate/infer the confidences
     label_encoder : bool, default=false
         Whether we need to apply encoding to the predictions.
+    model_metadata: Optional[dict], default=None
+        Additional metadata for the model.
 
     Attributes
     ----------
@@ -60,15 +88,15 @@ class FakedClassifier(BaseEstimator, ClassifierMixin):
     def __init__(self, oracle_X=None, predictions_=None, confidences_=None, oracle_index_=None,
                  simulate_n_features_in_=None,
                  predict_time_: float = 0, predict_proba_time_: float = 0, fit_time_: float = 0,
-                 simulate_time: bool = False, label_encoder=False):
+                 simulate_time: bool = False, label_encoder=False, model_metadata=None):
         self.simulate_time = simulate_time
         self.label_encoder = label_encoder
 
         self.fit_time_ = fit_time_
+        self.model_metadata = model_metadata
 
         # --- Init Parameter Validation
         if (oracle_X is not None) or (simulate_n_features_in_ is None):
-
             if confidences_ is None:
                 oracle_X, predictions_ = self._validate_simulation_data(oracle_X, predictions_, reset=True)
             elif predictions_ is None:
@@ -150,20 +178,22 @@ class FakedClassifier(BaseEstimator, ClassifierMixin):
         test_data_hash_indices = np.apply_along_axis(self._generate_index_from_row, 1, X)
 
         # Following Idea/Code From: https://www.statology.org/numpy-find-index-of-value/ to find index
+        #   This will set the values of duplicates to the first occurrences of the duplicate
+        #       (because that is the first index that is found by argsort).
         sorter = np.argsort(self.oracle_index_)
         prediction_indices = sorter[np.searchsorted(self.oracle_index_,
                                                     test_data_hash_indices, sorter=sorter)]
 
-        # Update labels similar to that in the meta model (FIXME, this wont work always...)
+        # Update labels similar to that in the meta model
         if self.label_encoder:
-            predictions = np.unique(self.predictions_, return_inverse=True)[1][prediction_indices]
+            predictions = self.le_.transform(self.predictions_[prediction_indices])
         else:
             predictions = self.predictions_[prediction_indices]
 
         if self.simulate_time:
             time.sleep(self.predict_time_)
 
-        return self.classes_[self.le_.transform(predictions)]
+        return predictions
 
     def predict_proba(self, X):
         """ Predicting with the fake classifier, that is, returning the previously stored confidences.
@@ -200,6 +230,8 @@ class FakedClassifier(BaseEstimator, ClassifierMixin):
         test_data_hash_indices = np.apply_along_axis(self._generate_index_from_row, 1, X)
 
         # Following Idea/Code From: https://www.statology.org/numpy-find-index-of-value/ to find index
+        #   This will set the values of duplicates to the first occurrences of the duplicate
+        #       (because that is the first index that is found by argsort).
         sorter = np.argsort(self.oracle_index_)
         confidences_indices = sorter[np.searchsorted(self.oracle_index_,
                                                      test_data_hash_indices, sorter=sorter)]
@@ -274,28 +306,32 @@ class FakedClassifier(BaseEstimator, ClassifierMixin):
 
     @staticmethod
     def _generate_index_from_row(x):
-        # Currently "just" a hash and not a full checksum algorithm... might need to change this
-        return hash(x.tobytes())  # Not consistent between runs (could not be stored), use hashlib for that
+        # See Class docu above why this works
+        return hash(tuple(x))
 
 
 # -- Additional Functions for FakedClassifiers Usage
 def _initialize_fake_models(test_base_model_train_X, test_base_model_train_y, val_base_model_train_X,
                             val_base_model_train_y, known_X, known_predictions, known_confidences, pre_fit_base_models,
-                            base_models_with_names, label_encoder, to_confidence_name):
+                            base_models_with_names, label_encoder, to_confidence_name, predictor_descriptions):
     # !WARNING!: Long text to explain to myself why the following works
     # Theoretically, we would need to build two separate base model sets for validation and test, because the base
     #   models might have been fitted differently for the validation predictions and test predictions.
-    #   (unless we take the validation data from the test predictions like in the no-validation-data case)
+    #   That is, in the case of a refit.
     # However, in our "lookup-table-like" fake base model, the fit method validates that the number of features
-    #   is correct. Which it must be for both cases in our setup. Otherwise it would be a bug.
-    #   Moreover, in classification it uses np.unique(y) to find the classes used to convert the predictions. This is,
-    #   however, always the same for val_y and test_y because we are always doing a stratified splits and a validation
-    #   split with only 1 class in it would be ill-defined. Hence, we can assume that np.unique(y) is identical for
-    #   val_fit and test_fit.
-    # In short, in our simulation the difference between data used to fit the base model for validation or test (outer
+    #   is correct. Which it must be for both cases in our setup. Otherwise, it would be a bug.
+    # Moreover, in classification it uses np.unique(y) to find the classes used to convert the predictions. This is,
+    #   however, almost always the same for val_y and test_y because we are always doing a stratified splits.
+
+    # TODO: fix this problem below
+    #   If refit + the validation data, with enough classes, it can happen that test_base_model_train_y has a different
+    #   number of classes than val_base_model_train_y, then a problem would exist. This is not yet supported / fixed.
+
+    # In any other case, the difference between data used to fit the base model for validation or test (outer
     #   evaluation) can be ignored. The base model can be (pre-)fitted on either validation or test data.
-    # It is only important that known_X contains all instances and known_predictions/confidences all output values
-    #   that would be needed during prediction for validation (ensemble fit) or outer evaluation (ensemble predict)
+    # It is only important that known_X contains all instances, known_predictions/confidences all output values
+    #   that would be needed during prediction for validation (ensemble fit) or outer evaluation (ensemble predict),
+    #   and the label encoder all possible classes.
 
     # -- Set values to one of validation or training data here
     classes_ = np.unique(test_base_model_train_y)
@@ -303,13 +339,16 @@ def _initialize_fake_models(test_base_model_train_X, test_base_model_train_y, va
     y_train = test_base_model_train_y
 
     # - Sanity Check to make sure we are not in an ill-defined split and everything described as above holds
-    #   (is technically a large overhead but my sanity is more important than the code's efficiency)
     if list(classes_) != list(np.unique(val_base_model_train_y)):
-        raise ValueError("Classes found in Validation and Test data used to fit the base model are not identical. "
-                         + "Something went wrong and I can not tell you where.")
+        val_classes = np.unique(val_base_model_train_y)
+        logger.info("Classes found in Validation and Test data used to fit the base model are not identical."
+                    + "\nClasses in test data: {}".format(list(classes_))
+                    + "\nClasses in val data: {}".format(list(val_classes)))
+        raise NotImplementedError("This can only occur when refit is true. This is not yet supported.")
 
     # -- Build fake base models
     faked_base_models = []
+    add_metadata = predictor_descriptions is not None
 
     for model_name in list(known_predictions):
         # Sort Predictions to the default predict_proba style and get predictions + confidences for a specific model
@@ -317,7 +356,8 @@ def _initialize_fake_models(test_base_model_train_X, test_base_model_train_y, va
         model_predictions = known_predictions[model_name]
 
         # Fit the FakedClassifier
-        fc = FakedClassifier(known_X, model_predictions, model_confidences, label_encoder=label_encoder)
+        fc = FakedClassifier(known_X, model_predictions, model_confidences, label_encoder=label_encoder,
+                             model_metadata=predictor_descriptions[model_name] if add_metadata else None)
 
         # -- Set fitted or not (sklearn vs. deslib)
         if pre_fit_base_models:
