@@ -8,6 +8,8 @@ from assembled.utils.logger import get_logger
 from typing import List, Tuple, Optional, Callable, Union
 from sklearn.model_selection import train_test_split
 
+from tables.exceptions import HDF5ExtError
+
 logger = get_logger(__file__)
 
 
@@ -28,13 +30,13 @@ class MetaTask:
         If True, we use pandas' sparse dtype to store data that is specific to a fold. This can drastically reduce
         the memory usage.
         Currently, due to a bug of pandas, only for confidence columns. FIXME fix this (make all labels numbers?)
-    use_hdf_file_format: bool, default=False
-        If True, we use the hdf file format to save and load the meta-dataset.
+    file_format: str in {"hdf", "csv", "feather"}, default="csv"
+        Determines which file format to use.
     """
 
-    def __init__(self, use_sparse_dtype: bool = True, use_hdf_file_format: bool = False):
+    def __init__(self, use_sparse_dtype: bool = True, file_format: str = "csv"):
         self.use_sparse_dtype = use_sparse_dtype
-        self.use_hdf_file_format = use_hdf_file_format
+        self.file_format = file_format
 
         # -- for dataset init
         self.dataset = None
@@ -69,6 +71,10 @@ class MetaTask:
         self.confidence_prefix = "confidence"
         self.fold_postfix = "fold"
         self.fold_predictor_prefix = "FP"
+        self.save_chunk_size = 1500
+        self._delayed_evaluation_load = False
+        self._file_load_path = None
+        self._custom_meta_data_container = dict()
 
         # -- Randomness (Experimental)
         self.random_int_seed_outer_folds = None
@@ -163,6 +169,10 @@ class MetaTask:
     def max_fold(self):
         return int(self.folds.max())
 
+    @property
+    def fold_indices(self):
+        return [int(f_i) for f_i in np.unique(self.folds)]
+
     # -- Properties related to meta-data
     @property
     def meta_data(self):
@@ -177,7 +187,7 @@ class MetaTask:
                 "cat_feature_names", "selection_constraints", "task_type", "predictor_corruptions_details",
                 "use_validation_data", "random_int_seed_outer_folds", "random_int_seed_inner_folds",
                 "folds", "fold_postfix", "fold_predictor_prefix", "validation_indices", "use_sparse_dtype",
-                "use_hdf_file_format"]
+                "file_format", "_custom_meta_data_container"]
 
     # -- Properties related to meta_dataset
     @property
@@ -212,8 +222,11 @@ class MetaTask:
 
     @property
     def yield_sparse_columns_chunks(self):
-        n = 100
+        n = self.save_chunk_size
         sparse_cols = self.sparse_columns
+
+        logger.info(
+            "Chunk Size: {} | #Columns: {} | Expected Chunks: {}".format(n, len(sparse_cols), len(sparse_cols) // n))
         for i in range(0, len(sparse_cols), n):
             yield sparse_cols[i:i + n]
 
@@ -521,26 +534,26 @@ class MetaTask:
 
         # -- Add normal prediction data
         if not fold_predictor:
-            self.predictions_and_confidences = self._get_prediction_data(predictions, confidences, predictor_name,
-                                                                         class_labels_to_use, conf_col_names,
-                                                                         self.predictions_and_confidences.copy())
+            self.predictions_and_confidences = pd.concat(
+                [self.predictions_and_confidences,
+                 self._get_prediction_data(predictions, confidences, predictor_name, class_labels_to_use,
+                                           conf_col_names, self.predictions_and_confidences.index)], axis=1)
         else:
             other_indices, predicted_on_indices = self.get_indices_for_fold(fold_predictor_idx, return_indices=True)
 
-            self.predictions_and_confidences = self._get_prediction_data(predictions, confidences, predictor_name,
-                                                                         class_labels_to_use, conf_col_names,
-                                                                         self.predictions_and_confidences.copy(),
-                                                                         fold_data=True,
-                                                                         fold_data_arguments={
-                                                                             "non_used_indices": other_indices,
-                                                                             "used_indices": predicted_on_indices
-                                                                         })
+            self.predictions_and_confidences = pd.concat(
+                [self.predictions_and_confidences,
+                 self._get_prediction_data(predictions, confidences, predictor_name, class_labels_to_use,
+                                           conf_col_names, self.predictions_and_confidences.index,
+                                           fold_data=True, fold_data_arguments={
+                         "non_used_indices": other_indices,
+                         "used_indices": predicted_on_indices
+                     })], axis=1)
 
         # -- Add validation data
         if validation_data is not None:
             # (Re-)Mark validation data usage
             self.use_validation_data = True
-            tmp_val_data = self.validation_predictions_and_confidences.copy()
 
             for fold_idx, val_preds, val_confs, val_indices in validation_data:
                 # -- Fold preliminaries
@@ -570,22 +583,23 @@ class MetaTask:
                 non_val_indices = np.append(non_val_indices, additional_non_val_indices)
 
                 # Get fold prediction data
-                tmp_val_data = self._get_prediction_data(val_preds, val_confs, save_name,
-                                                         class_labels_to_use, val_conf_col_names,
-                                                         tmp_val_data,
-                                                         fold_data=True,
-                                                         fold_data_arguments={
-                                                             "non_used_indices": non_val_indices,
-                                                             "used_indices": val_indices
-                                                         })
+                self.validation_predictions_and_confidences = pd.concat(
+                    [self.validation_predictions_and_confidences,
+                     self._get_prediction_data(val_preds, val_confs, save_name, class_labels_to_use, val_conf_col_names,
+                                               self.validation_predictions_and_confidences.index,
+                                               fold_data=True, fold_data_arguments={
+                             "non_used_indices": non_val_indices,
+                             "used_indices": val_indices
+                         })],
+                    axis=1)
 
             # -- Post-processing of fold data
             # Re-order validation data to have identical order at all times
-            self.validation_predictions_and_confidences = tmp_val_data[self.validation_predictions_columns +
-                                                                       self.validation_confidences_columns]
+            self.validation_predictions_and_confidences = self.validation_predictions_and_confidences[
+                self.validation_predictions_columns + self.validation_confidences_columns]
 
     def _get_prediction_data(self, predictions, confidences, predictor_name, class_labels_to_use, conf_col_names,
-                             current_prediction_data, fold_data: bool = False,
+                             current_prediction_data_index, fold_data: bool = False,
                              fold_data_arguments: dict = None):
         """Checks are formats prediction data where needed. Return contacted prediction data.
 
@@ -652,12 +666,11 @@ class MetaTask:
             pred_data = pred_data.astype(dtype_per_col)
 
         # -- Concat and Verify integrity of index
-        tmp_data = pd.concat([current_prediction_data, pred_data], axis=1).reset_index()
-        if sum(tmp_data.index != tmp_data["index"]) != 0:
+        pred_data = pred_data.reset_index()
+        if (current_prediction_data_index.size != 0) and sum(current_prediction_data_index != pred_data["index"]) != 0:
             raise ValueError("Something went wrong with the index of the predictions data!")
-        tmp_predictions_and_confidences = tmp_data.drop(columns=["index"])
 
-        return tmp_predictions_and_confidences
+        return pred_data.drop(columns=["index"])
 
     # - Loading Side Information
     def read_selection_constraints(self, selection_constraints):
@@ -725,26 +738,7 @@ class MetaTask:
         output_dir:  str
             Directory in which the .json and .csv files for the metatask shall be stored.
         """
-        # -- Store the predictions together with the dataset in one file
-        if self.use_hdf_file_format:
-            logger.info("Start Saving to HDF File...")
-            file_path_hdf = os.path.join(output_dir, "metatask_{}.hdf".format(self.openml_task_id))
-            self.meta_dataset[self.dense_columns].to_hdf(file_path_hdf, str(self.openml_task_id), mode="w",
-                                                         format="table")
-
-            # Go in chunks over the columns
-            for chunk_idx, to_save_cols in enumerate(self.yield_sparse_columns_chunks, 1):
-                logger.info("Saving Column Chunks {}".format(chunk_idx))
-                # Get part of dataset to save and change its type
-                dtypes_dict = self.get_save_dtypes_for_columns(to_save_cols)
-                self.meta_dataset[to_save_cols].apply(lambda x: x.values.to_dense()).astype(
-                    dtypes_dict).to_hdf(file_path_hdf, str(self.openml_task_id) + str(to_save_cols), mode="r+",
-                                        format="table", append=True)
-
-        else:
-            logger.info("Start Saving to CSV File...")
-            file_path_csv = os.path.join(output_dir, "metatask_{}.csv".format(self.openml_task_id))
-            self.meta_dataset.to_csv(file_path_csv, sep=",", header=True, index=False)
+        self.store_meta_dataset(output_dir)
 
         # -- Store Meta data in its onw file
         logger.info("Start Saving to JSON File...")
@@ -760,6 +754,92 @@ class MetaTask:
 
         logger.info("Finished Saving Metatask to Files.")
 
+    def store_meta_dataset(self, output_dir):
+        out_f = self.file_format
+
+        # -- Store the predictions together with the dataset in one file
+        if out_f == "csv":
+            logger.info("Start Saving to CSV File...")
+            file_path_csv = os.path.join(output_dir, "metatask_{}.csv".format(self.openml_task_id))
+            self.meta_dataset.to_csv(file_path_csv, sep=",", header=True, index=False)
+
+        elif out_f == "feather":
+            logger.info("Start Saving to Feather Files...")
+            save_dir = os.path.join(output_dir, "metatask_{}".format(self.openml_task_id))
+            os.makedirs(save_dir, exist_ok=True)
+
+            logger.info("Start Saving Dense columns to feather File...")
+            dense_p = os.path.join(save_dir, "dense_metatask_{}.feather".format(self.openml_task_id))
+            self.meta_dataset[self.dense_columns].to_feather(dense_p)
+
+            logger.info("Start Saving Sparse columns to feather File...")
+            for chunk_idx, to_save_cols in enumerate(self.yield_sparse_columns_chunks, 1):
+                logger.info("Saving Column Chunk {}".format(chunk_idx))
+                dtypes_dict = self.get_save_dtypes_for_columns(to_save_cols)
+                s_chunk_p = os.path.join(save_dir,
+                                         "sparse_metatask_{}_{}.feather".format(self.openml_task_id, chunk_idx))
+                self.meta_dataset[to_save_cols].apply(lambda x: x.values.to_dense()
+                                                      ).astype(dtypes_dict).to_feather(s_chunk_p)
+        elif out_f == "hdf":
+            logger.info("Start Saving to HDF as Split Files...")
+            save_p = os.path.join(output_dir, "metatask_{}.hdf".format(self.openml_task_id))
+
+            logger.info("Start Saving Dataset...")
+            try:
+                self.dataset.to_hdf(save_p, "dataset", mode="w", format="table")
+            except HDF5ExtError:
+                logger.info("Dataset is too large to save in one HDF Table. Need to split it up.")
+
+                # Save in chunks
+                self._save_to_hdf_splits(self.dataset, save_p, "dataset")
+
+                # Clean up
+                with pd.HDFStore(save_p, mode="r+") as store:
+                    store.remove("dataset")
+
+            logger.info("Start Saving Prediction Data...")
+            fold_indices = [int(f_i) for f_i in np.unique(self.folds)]
+            for fold_index in fold_indices:
+                logger.info("Save for Fold: {}".format(fold_index))
+                f_pred = self.get_predictors_for_fold(fold_index)
+
+                # Enable HDF for incomplete fold data or empty metatasks.
+                if not f_pred:
+                    logger.info("No prediction Data for this Fold... skip.")
+                else:
+                    logger.info("Save Test Prediction Data...")
+                    f_cols = self.get_pred_and_conf_cols(f_pred)
+                    _, f_test_i = self.get_indices_for_fold(fold_index)
+
+                    self._save_to_hdf_splits(self.predictions_and_confidences.loc[f_test_i, f_cols],
+                                             save_p, "t_p_d_{}".format(fold_index))
+
+                if self.use_validation_data:
+                    logger.info("Save Validation Prediction Data...")
+                    f_cols = self.get_validation_predictions_columns(f_pred) \
+                             + self.get_validation_confidences_columns(f_pred)
+                    f_val_i = self.validation_indices[fold_index]
+                    self._save_to_hdf_splits(self.validation_predictions_and_confidences.loc[f_val_i, f_cols],
+                                             save_p, "v_p_d_{}".format(fold_index))
+
+        else:
+            raise ValueError("Unknown Output Format: {}".format(out_f))
+
+    def _save_to_hdf_splits(self, df_to_save, out_path, hdf_key):
+        sparse_catcher = lambda x: x.sparse.to_dense() if pd.api.types.is_sparse(x) else x
+
+        # Avoid "object header message is too large" bug/feature of HDF
+        cols = list(df_to_save)  # This is equivalent to the order produced before.
+        n = self.save_chunk_size
+        chunk_indices = []
+        for i in range(0, len(cols), n):
+            chunk_indices.append(i)
+            col_chunk = cols[i:i + n]
+            df_to_save[col_chunk].apply(sparse_catcher).to_hdf(out_path, hdf_key + f"_{i}", mode="r+", format="table")
+
+        # Save Chunk metadata
+        pd.Series(chunk_indices).to_hdf(out_path, hdf_key + "_md", mode="r+", format="table")
+
     def to_sharable_prediction_data(self, output_dir: str = ""):
         """Store Metatasks without dataset (e.g., all data but self.dataset's rows)"""
 
@@ -771,7 +851,8 @@ class MetaTask:
         self.to_files(output_dir)
 
     # - Metatask from files
-    def read_metatask_from_files(self, input_dir: str, openml_task_id: int, read_wo_dataset: bool = False):
+    def read_metatask_from_files(self, input_dir: str, openml_task_id: int, read_wo_dataset: bool = False,
+                                 delayed_evaluation_load: bool = False):
         """ Build a metatask using data from files
 
         Parameters
@@ -783,17 +864,21 @@ class MetaTask:
         read_wo_dataset: bool, default=False
             If the function is used to only read the meta data and prediction data from the files.
             Needed to determine which sanity checks to run.
+        delayed_evaluation_load: bool, default=False
+            If true, the prediction data will be only loaded  once needed for the evaluation of a fold.
+            Afterwards, the prediction data is also removed again. In other words, it is loaded only for a fold.
+            That is, if fold_split is called.
+            Only supported for file formats in {"hdf"} for now.
         """
-        # -- Read data
-        file_path_json = os.path.join(input_dir, "metatask_{}.json".format(openml_task_id))
 
         # - Read meta data
+        file_path_json = os.path.join(input_dir, "metatask_{}.json".format(openml_task_id))
         with open(file_path_json) as json_file:
             meta_data = json.load(json_file)
 
         # -- Init Meta Data
         for md_k in meta_data.keys():
-            setattr(self, md_k, meta_data[md_k])
+            setattr(self, md_k, meta_data[md_k])  # Supports backwards compatibility
 
         # Post process special cases (setting arrays to arrays)
         self.folds = np.array(self.folds)
@@ -803,9 +888,23 @@ class MetaTask:
         self.missing_metadata_in_file = [x for x in self.meta_data_keys if x not in meta_data.keys()]
         del meta_data
 
+        # -- Get and parse dataset
+        if delayed_evaluation_load and (self.file_format not in ["hdf"]):
+            raise NotImplementedError(
+                "Delayed evaluation load is not yet supported for file format {}".format(self.file_format))
+        self._delayed_evaluation_load = delayed_evaluation_load
+
+        self.read_meta_dataset(input_dir, openml_task_id)
+
+        if not read_wo_dataset:
+            self._dataset_sanity_checks()
+
+    def read_meta_dataset(self, input_dir, openml_task_id):
+        out_f = self.file_format
+
         # -- Read Dataset
-        if not self.use_hdf_file_format:
-            file_path_csv = os.path.join(input_dir, "metatask_{}.csv".format(openml_task_id))
+        if out_f == "csv":
+            load_path = os.path.join(input_dir, "metatask_{}.csv".format(openml_task_id))
 
             dtypes_to_read = {col_name: 'category' for col_name in self.cat_feature_names}
 
@@ -822,39 +921,157 @@ class MetaTask:
             # Init Dataset and Handle sparse dtypes
             if self.use_sparse_dtype:
                 # Read the non sparse part
-                meta_dataset = pd.read_csv(file_path_csv, dtype=dtypes_to_read, usecols=self.dense_columns)
+                meta_dataset = pd.read_csv(load_path, dtype=dtypes_to_read, usecols=self.dense_columns)
 
                 # Go in chunks over the columns
                 for to_load_cols in self.yield_sparse_columns_chunks:
-                    tmp_md = pd.read_csv(file_path_csv, usecols=to_load_cols)[to_load_cols]
+                    tmp_md = pd.read_csv(load_path, usecols=to_load_cols)[to_load_cols]
                     dtype_per_col = self.get_load_dtypes_for_columns(to_load_cols)
                     meta_dataset[to_load_cols] = tmp_md.astype(dtype_per_col)
 
             else:
-                meta_dataset = pd.read_csv(file_path_csv, dtype=dtypes_to_read)
-        else:
-            file_path_hdf = os.path.join(input_dir, "metatask_{}.hdf".format(openml_task_id))
-            meta_dataset = pd.read_hdf(file_path_hdf, key=str(openml_task_id))
+                meta_dataset = pd.read_csv(load_path, dtype=dtypes_to_read)
 
-            # Go in chunks over the columns
-            for to_load_cols in self.yield_sparse_columns_chunks:
-                # Get part of dataset to save and change its type
+        elif out_f == "feather":
+            logger.info("Start Loading Feather File...")
+            save_dir = os.path.join(input_dir, "metatask_{}".format(self.openml_task_id))
+
+            logger.info("Start Loading Dense columns...")
+            dense_p = os.path.join(save_dir, "dense_metatask_{}.feather".format(self.openml_task_id))
+            meta_dataset = pd.read_feather(dense_p)
+
+            logger.info("Start Loading Sparse columns...")
+            for chunk_idx, to_load_cols in enumerate(self.yield_sparse_columns_chunks, 1):
                 dtypes_dict = self.get_load_dtypes_for_columns(to_load_cols)
-                meta_dataset[to_load_cols] = pd.read_hdf(file_path_hdf, key=str(openml_task_id)
-                                                                            + str(to_load_cols)).astype(dtypes_dict)
+                s_chunk_p = os.path.join(save_dir,
+                                         "sparse_metatask_{}_{}.feather".format(self.openml_task_id, chunk_idx))
+                meta_dataset[to_load_cols] = pd.read_feather(s_chunk_p).astype(dtypes_dict)
 
+        elif out_f == "hdf":
+            logger.info("Start Loading from HDF Split Files...")
+            load_p = os.path.join(input_dir, "metatask_{}.{}".format(openml_task_id, out_f))
+
+            logger.info("Start Loading Dataset...")
+            # Determine save type of dataset
+            with pd.HDFStore(load_p, mode="r") as store:
+                one_table_dataset = "dataset" in store
+
+            if one_table_dataset:
+                meta_dataset = pd.read_hdf(load_p, "dataset")
+            else:
+                # self.folds gives us the number of instances
+                meta_dataset = self._read_dataset_from_hdf_splits(load_p)
+
+            if not self._delayed_evaluation_load:
+                if self.predictors:  # workaround for empty prediction data, FIXME
+                    logger.info("Start Loading Prediction Data...")
+                    # Iteration
+                    for fold_index in self.fold_indices:
+                        meta_dataset = pd.concat(
+                            [meta_dataset, self._read_hdf_for_fold(fold_index, meta_dataset.index, load_p)],
+                            axis=1)
+            else:
+                self._file_load_path = load_p
+
+        else:
+            raise ValueError("Unknown file format: {}".format(out_f))
+
+        # Always save dataset
         self.dataset = meta_dataset[self.feature_names + [self.target_name]]
-        self.predictions_and_confidences = meta_dataset[self.pred_and_conf_cols]
+        self._fill_prediction_data(meta_dataset)
 
-        if self.validation_predictions_columns or self.validation_confidences_columns:
+    def _fill_prediction_data(self, meta_dataset):
+        # Decided what prediction data to save
+        if (not self._delayed_evaluation_load) and (self.pred_and_conf_cols):
+            self.predictions_and_confidences = meta_dataset[self.pred_and_conf_cols]
+        else:
+            # Fallback
+            self.predictions_and_confidences = pd.DataFrame()
+
+        if (not self._delayed_evaluation_load) and (self.validation_predictions_columns
+                                                    or self.validation_confidences_columns):
             self.validation_predictions_and_confidences = meta_dataset[self.validation_predictions_columns +
                                                                        self.validation_confidences_columns]
         else:
+            # Fallback
             self.validation_predictions_and_confidences = pd.DataFrame()
-        del meta_dataset
 
-        if not read_wo_dataset:
-            self._dataset_sanity_checks()
+    def _read_hdf_for_fold(self, fold_index, meta_dataset_indices, path_to_hdf_file):
+        logger.info("Load for Fold: {}".format(fold_index))
+
+        logger.info("Load Test Prediction Data...")
+        pred_and_confs = self._read_from_hdf_splits(meta_dataset_indices, path_to_hdf_file,
+                                                    "t_p_d_{}".format(fold_index))
+
+        if self.use_validation_data:
+            logger.info("Load Validation Prediction Data...")
+            val_pred_and_confs = self._read_from_hdf_splits(meta_dataset_indices, path_to_hdf_file,
+                                                            "v_p_d_{}".format(fold_index))
+            return pd.concat([pred_and_confs, val_pred_and_confs], axis=1)
+
+        else:
+            return pred_and_confs
+
+    def _read_dataset_from_hdf_splits(self, in_path):
+        chunk_indices = pd.read_hdf(in_path, "dataset_md").tolist()
+
+        overall_df = None
+
+        for i in chunk_indices:
+            # Load Data
+            tmp_pd = pd.read_hdf(in_path, f"dataset_{i}")
+
+            if overall_df is None:
+                overall_df = tmp_pd
+            else:
+                overall_df = pd.concat([overall_df, tmp_pd], axis=1)
+
+        return overall_df
+
+    def _read_from_hdf_splits(self, meta_dataset_indices, in_path, hdf_key):
+        chunk_indices = pd.read_hdf(in_path, hdf_key + "_md").tolist()
+        sparse_cols = set(self.sparse_columns)
+        tmp_df_store = pd.DataFrame(index=meta_dataset_indices)
+
+        for i in chunk_indices:
+            # Load Data
+            tmp_pd = pd.read_hdf(in_path, hdf_key + f"_{i}")
+            indices = tmp_pd.index
+
+            # Determine dtype
+            all_cols_for_type = {}
+            for col_name, type_d in zip(tmp_pd.columns, tmp_pd.dtypes):
+                load_dtype = type_d
+                final_dtype = type_d
+
+                if col_name in sparse_cols:
+                    final_dtype = pd.SparseDtype("float64", np.nan)
+
+                use_key = (load_dtype, final_dtype)
+
+                if use_key not in all_cols_for_type:
+                    all_cols_for_type[use_key] = []
+
+                all_cols_for_type[use_key].append(col_name)
+
+            # Handle insert into existing data
+            for (load_dt, final_dt), col_names in all_cols_for_type.items():
+                tmp_df_store = pd.concat([tmp_df_store,
+                                          pd.DataFrame(index=meta_dataset_indices, dtype=load_dt, columns=col_names)],
+                                         axis=1)
+                tmp_df_store.loc[indices, col_names] = tmp_pd[col_names]
+                tmp_df_store = tmp_df_store.astype({c: final_dt for c in col_names})
+
+        return tmp_df_store
+
+    def read_prediction_data_for_fold(self, fold_index):
+        """ Only Read the Prediction data (test and validation) for a specific fold"""
+        if self.file_format == "hdf":
+            pred_data = self._read_hdf_for_fold(fold_index, self.dataset.index, self._file_load_path)
+        else:
+            raise ValueError("Unsupported File Format for Fold Read.")
+
+        return pred_data
 
     def from_sharable_prediction_data(self, input_dir: str, openml_task_id: int, dataset: pd.DataFrame):
         # Get Shared Data
@@ -1044,11 +1261,21 @@ class MetaTask:
 
         return train_indices, test_indices
 
-    def fold_split(self, return_fold_index=False):
+    def fold_split(self, return_fold_index=False, folds_to_run=None):
         # Return a fold's split copy of metadataset
-        yield_copy = self.meta_dataset.copy()
+
+        # Only yield full copy if needed
+        if not self._delayed_evaluation_load:
+            yield_copy = self.meta_dataset.copy()
 
         for i in range(self.max_fold + 1):
+            # Skip not needed folds
+            if (folds_to_run is not None) and (i not in folds_to_run):
+                continue
+
+            if self._delayed_evaluation_load:
+                yield_copy = pd.concat([self.dataset, self.read_prediction_data_for_fold(i)], axis=1)
+
             train_indices, test_indices = self.get_indices_for_fold(i)
 
             if return_fold_index:
@@ -1157,11 +1384,7 @@ class MetaTask:
             Confidences of each base model on the fold's test data
         """
 
-        for idx, train_metadata, test_metadata in self.fold_split(return_fold_index=True):
-            # -- Check if Fold is wanted, if not skip
-            if (folds_to_run is not None) and (idx not in folds_to_run):
-                continue
-
+        for idx, train_metadata, test_metadata in self.fold_split(return_fold_index=True, folds_to_run=folds_to_run):
             X_train, y_train, _, _, val_base_predictions, val_base_confidences = self.split_meta_dataset(train_metadata,
                                                                                                          fold_idx=idx)
             X_test, y_test, test_base_predictions, test_base_confidences, _, _ = self.split_meta_dataset(test_metadata,
@@ -1174,7 +1397,8 @@ class MetaTask:
     def _exp_yield_evaluation_data_across_folds(self, meta_train_test_split_fraction,
                                                 meta_train_test_split_random_state,
                                                 pre_fit_base_models, base_models_with_names, label_encoder,
-                                                preprocessor, include_test_data=False):
+                                                preprocessor, include_test_data=False,
+                                                store_metadata_in_fake_base_model=False):
         from assembled.compatibility.faked_classifier import _initialize_fake_models
 
         # FIXME, extend to validation data usage?
@@ -1193,7 +1417,8 @@ class MetaTask:
 
             base_models = _initialize_fake_models(X_train, y_train, X_train, y_train, X_test, test_base_predictions,
                                                   test_base_confidences, pre_fit_base_models, base_models_with_names,
-                                                  label_encoder, self.to_confidence_name)
+                                                  label_encoder, self.to_confidence_name,
+                                                  self.predictor_descriptions if store_metadata_in_fake_base_model else None)
 
             if include_test_data:
                 assert_meta_train_pred, assert_meta_test_pred, assert_meta_train_conf, \
