@@ -6,6 +6,8 @@ from multiprocessing import Process, Queue
 def isolate_function(func, *i_args, **i_kwargs):
     """The Processify Decorate adapted with minor changes to be a callable wrapper
 
+    We also modified it such that it returns the technique args because of random state.
+
     Taken from and all Credits to https://gist.github.com/schlamar/2311116
     """
 
@@ -19,7 +21,8 @@ def isolate_function(func, *i_args, **i_kwargs):
         else:
             error = None
 
-        q.put((ret, error))
+        _, _, technique_args, *_ = args
+        q.put((ret, error, technique_args))
 
     # register original function with different name
     # in sys.modules so it is pickable
@@ -29,10 +32,10 @@ def isolate_function(func, *i_args, **i_kwargs):
     q = Queue()
     p = Process(target=process_func, args=tuple([q] + list(i_args)), kwargs=i_kwargs)
     p.start()
-    ret, error = q.get()
+    ret, error, technique_args = q.get()
     p.join()
 
     if error:
         raise RuntimeError("Error in Isolate Subprocess, see previous traceback.")
 
-    return ret
+    return ret, technique_args
